# Author                 Version      Date
# -----------------------------------------------
# S M Crawford (SAA0)    0.1          20 Feb 2008
#
# slotphottools contains the algorithms to perform 
# photometry on slotmode data.  

from pyraf import iraf 
import saltprint, saltkey, saltio, saltstat
from numpy import *


# -----------------------------------------------------------
# Background subtraction 

def subbackground(image, sig, bin, order, iter, type, log):
    """Calculate and subtract a global background from an image
       using a number of different methods.  This is assuming that
       slotmode data is being used.
	
       return data array, status
    """

    status=0

    #return the same image if no background subtraction is required
    if type=='none': return image, status
 
    #calculate the image statistics
    if size(image)>0:
	mean, median, stddev=saltstat.iterstat(image,sig,iter)
    if stddev==0: status=1

    #Create a row median image.  This will be useful for a number of purposes
    if status==0:
       image_mrow,status=med_row_image(image, log)

    #Replace bright objects in the frame with the median value of that row
    if status==0:
     mask = ((image -mean) < sig*stddev)*(image -mean > -sig*stddev)
     image_sig= image * mask + (1-mask)*image_mrow

    if status==0 and type == 'median-row':
     image_back=image*0.0+image_mrow

    #Median smooth the image
    if status==0 and type == 'median':
     image_back,status=median_image(image_sig, bin, log)

    #make a fit to the surface
    if status==0 and type=='surf': 
        image_back=image.copy()

    if status==0  and type=='both':
     image_surf=image.copy()
     for x in range(len(image[0])):
	y=arange(len(image))
	my=((image[:,x]-mean) < sig*stddev)*(image[:,x] -mean > -sig*stddev)
	cy=compress(my,y)
	cim=compress(my,image_med[:,x])
	coef=polyfit(cy,cim,order)
	image_surf[:,x]=polyval(coef,y)

    #subtract the fit
    if status==0:
      image = image-image_back

    #return the image   
    return image,status

# -----------------------------------------------------------
# Median filter an array

def median_image(arr, bin, logfile):
    """Median smooth an image with a filter size set by bin

    returns arr , status
    """
    from scipy.ndimage.filters import median_filter
    status=0
    try:
       arr=median_filter(arr,size=(bin,bin))
    except:
       message="SLOTBACKGROUND--ERROR:  Could not median filter image"
       status=saltio.err(logfile, message)

    return arr,status

# -----------------------------------------------------------
# Median filter an array

def med_row_image(arr, logfile):
    """Calculate the median value for each row in the array

        returns arr , status
    """
    status=0
    nrows=len(arr)
    
    try:
       marr=zeros((nrows,1),dtype=float)
       for i in range(nrows):
    	marr[i]=median(arr[i])
    except Exception, e:
       message="SLOTBACKGROUND--ERROR:  Could not calculate median row values " + e
       status=saltprint.err(logfile, message)
    return marr,status




